import pytest
import unittest
from unittest.mock import Mock
from pynonymizer.strategy.update_column import (
    UpdateColumnStrategyTypes,
    EmptyUpdateColumnStrategy,
    UniqueLoginUpdateColumnStrategy,
    UniqueEmailUpdateColumnStrategy,
    LiteralUpdateColumnStrategy,
    FakeUpdateColumnStrategy
)

"""
Tests for making sure strategies are built in the way that we need them to be
Particularly, behaviour surrounding args and arg validation, and strategy-wide constraints (like the `where` option!)
"""

def test_empty():
    empty = EmptyUpdateColumnStrategy(column_name="empty_column")
    assert empty.strategy_type == UpdateColumnStrategyTypes.EMPTY
    assert empty.where_condition == None


def test_empty_where():
    empty_where = EmptyUpdateColumnStrategy(column_name="empty_column", where="`cheese` = 'gouda'")

    assert empty_where.strategy_type == UpdateColumnStrategyTypes.EMPTY
    assert empty_where.where_condition == "`cheese` = 'gouda'"


def test_ulogin():
    ulogin = UniqueLoginUpdateColumnStrategy(column_name="column_name")

    assert ulogin.strategy_type == UpdateColumnStrategyTypes.UNIQUE_LOGIN
    assert ulogin.where_condition == None


def test_ulogin_where():
    ulogin_where = UniqueLoginUpdateColumnStrategy(column_name="column_name",where="`cheese` = 'edam'")

    assert ulogin_where.strategy_type == UpdateColumnStrategyTypes.UNIQUE_LOGIN
    assert ulogin_where.where_condition == "`cheese` = 'edam'"


def test_uemail():
    uemail = UniqueEmailUpdateColumnStrategy(column_name="column_name")

    assert uemail.strategy_type == UpdateColumnStrategyTypes.UNIQUE_EMAIL
    assert uemail.where_condition == None


def test_uemail_where():
    uemail = UniqueEmailUpdateColumnStrategy(column_name="column_name",where="`cheese` = 'gouda'")

    assert uemail.strategy_type == UpdateColumnStrategyTypes.UNIQUE_EMAIL
    assert uemail.where_condition == "`cheese` = 'gouda'"


def test_literal():
    literal = LiteralUpdateColumnStrategy(column_name="column_name",value="52")

    assert literal.strategy_type == UpdateColumnStrategyTypes.LITERAL
    assert literal.where_condition == None
    assert literal.value == "52"


def test_literal_where():
    literal_where = LiteralUpdateColumnStrategy(column_name="column_name",value="RAND()", where="cake = 'death' OR gravel > 3")

    assert literal_where.strategy_type == UpdateColumnStrategyTypes.LITERAL
    assert literal_where.where_condition == "cake = 'death' OR gravel > 3"
    assert literal_where.value == "RAND()"


def test_fake_update():
    fake_column_generator = Mock()
    fake_update = FakeUpdateColumnStrategy(column_name="column_name",fake_column_generator=fake_column_generator, fake_type="company_email")

    assert fake_update.strategy_type == UpdateColumnStrategyTypes.FAKE_UPDATE
    assert fake_update.value == fake_column_generator.get_value("company_email")
    assert fake_update.data_type == fake_column_generator.get_data_type("company_email")
    assert fake_update.fake_type == "company_email"


def test_fake_update_where():
    fake_column_generator = Mock()
    fake_update_where = FakeUpdateColumnStrategy(column_name="column_name",fake_column_generator=fake_column_generator, fake_type="company_email", where="chickens = 1")

    assert fake_update_where.strategy_type == UpdateColumnStrategyTypes.FAKE_UPDATE
    assert fake_update_where.value == fake_column_generator.get_value("company_email")
    assert fake_update_where.data_type == fake_column_generator.get_data_type("company_email")
    assert fake_update_where.fake_type == "company_email"
    assert fake_update_where.where_condition == "chickens = 1"




import unittest
import pytest
import os
from unittest.mock import Mock, patch, MagicMock, call, mock_open
from pynonymizer.database.mysql import MySqlProvider
from pynonymizer.strategy.database import DatabaseStrategy
from pynonymizer.strategy.table import TruncateTableStrategy
from pynonymizer.database.exceptions import UnsupportedTableStrategyError
from tests.helpers import list_rindex


@patch("pynonymizer.database.mysql.execution", autospec=True)
def test_init_runners_correctly(execution):
    """
    Test the provider inits dependencies with the correct database information
    """
    MySqlProvider("1.2.3.4", "root", "password", "db_name")
    execution.MySqlCmdRunner.assert_called_once_with("1.2.3.4", "root", "password", "db_name")
    execution.MySqlDumpRunner.assert_called_once_with("1.2.3.4", "root", "password", "db_name")


@patch("pynonymizer.database.mysql.execution", autospec=True)
@patch("pynonymizer.database.mysql.query_factory", autospec=True)
def test_create_database(query_factory, execution):
    provider = MySqlProvider("1.2.3.4", "root", "password", "db_name")
    provider.create_database()

    # assert that the query factory is called with the db name, and the output is passed to the execute runner
    query_factory.get_create_database.assert_called_once_with("db_name")
    execution.MySqlCmdRunner.return_value.execute.assert_called_once_with(query_factory.get_create_database.return_value)


@patch("pynonymizer.database.mysql.execution", autospec=True)
@patch("pynonymizer.database.mysql.query_factory", autospec=True)
def test_drop_database(query_factory, execution):
    provider = MySqlProvider("1.2.3.4", "root", "password", "db_name")
    provider.drop_database()

    # assert that the query factory is called with the db name, and the output is passed to the execute runner
    query_factory.get_drop_database.assert_called_once_with("db_name")
    execution.MySqlCmdRunner.return_value.execute.assert_called_once_with(query_factory.get_drop_database.return_value)


@patch("pynonymizer.database.mysql.execution", autospec=True)
@patch("pynonymizer.database.mysql.query_factory", autospec=True)
def test_connection(query_factory, execution):
    # test_connection should return the cmd runner's test output (bool)
    provider = MySqlProvider("1.2.3.4", "root", "password", "db_name")
    assert provider.test_connection() == execution.MySqlCmdRunner.return_value.test.return_value


@patch("pynonymizer.database.mysql.execution", autospec=True)
@patch("pynonymizer.database.mysql.query_factory", autospec=True)
@patch("pynonymizer.database.mysql.resolve_input")
def test_restore_database(resolve_input, query_factory, execution):
    provider = MySqlProvider("1.2.3.4", "root", "password", "db_name")
    rand_data = bytes( os.urandom(8193) )
    mock_input = Mock(get_size=Mock(return_value=8193), open=mock_open(read_data=rand_data))
    resolve_input.return_value = mock_input

    provider.restore_database("testfile.bbs")

    # ask for size
    mock_input.get_size.assert_called()
    # open input and read at least once
    mock_input.open.assert_called()
    mock_input.open.return_value.read.assert_called()
    # open, write and flush at least once
    execution.MySqlCmdRunner.return_value.open_batch_processor.assert_called_once_with()
    execution.MySqlCmdRunner.return_value.open_batch_processor.return_value.write.assert_called()
    execution.MySqlCmdRunner.return_value.open_batch_processor.return_value.flush.assert_called()


@patch("pynonymizer.database.mysql.execution", autospec=True)
@patch("pynonymizer.database.mysql.query_factory", autospec=True)
@patch("pynonymizer.database.mysql.resolve_output")
def test_dump_database(resolve_output, query_factory, execution):
    provider = MySqlProvider("1.2.3.4", "root", "password", "db_name")
    rand_data = bytes( os.urandom(8192) )
    mock_output = Mock(open=mock_open())
    resolve_output.return_value = mock_output

    execution.MySqlDumpRunner.return_value.open_dumper.return_value = MagicMock(read=MagicMock(side_effect=[rand_data, b""]))

    provider.dump_database("testfile.bbs")

    # open output and write at least once
    mock_output.open.assert_called()
    mock_output.open.return_value.write.assert_called()

    # open dumper and read at least once
    execution.MySqlDumpRunner.return_value.open_dumper.assert_called_once_with()
    execution.MySqlDumpRunner.return_value.open_dumper.return_value.read.assert_called()


@patch("pynonymizer.database.mysql.execution", autospec=True)
@patch("pynonymizer.database.mysql.query_factory", autospec=True)
def test_anonymize_database_unsupported_table_strategy(query_factory, execution):
    with pytest.raises(UnsupportedTableStrategyError) as e_info:
        provider = MySqlProvider("1.2.3.4", "root", "password", "db_name")
        database_strategy = DatabaseStrategy([
                Mock(table_name="table1", strategy_type="DEFINITELY_NOT_A_SUPPORTED_STRATEGY_TYPE"),
        ])
        provider.anonymize_database(database_strategy)


@patch("pynonymizer.database.mysql.execution", autospec=True)
@patch("pynonymizer.database.mysql.query_factory", autospec=True)
def test_before_script_run(query_factory, execution):
    manager = Mock()
    manager.attach_mock(execution, "execution")

    provider = MySqlProvider("1.2.3.4", "root", "password", "db_name")
    database_strategy = DatabaseStrategy(table_strategies=[
            TruncateTableStrategy("table1")
        ],
        before_scripts=["SELECT `before` FROM `before_table`;"]
    )

    provider.anonymize_database(database_strategy)

    truncate_call_index = manager.mock_calls.index(call.execution.MySqlCmdRunner().db_execute(query_factory.get_truncate_table.return_value))
    before_script_call_index = manager.mock_calls.index( call.execution.MySqlCmdRunner().db_execute("SELECT `before` FROM `before_table`;") )

    # before script should be before any anonymization happens (truncate)
    assert before_script_call_index < truncate_call_index


@patch("pynonymizer.database.mysql.execution", autospec=True)
@patch("pynonymizer.database.mysql.query_factory", autospec=True)
def test_after_script_run(query_factory,execution):
    manager = Mock()
    manager.attach_mock(execution, "execution")

    provider = MySqlProvider("1.2.3.4", "root", "password", "db_name")
    database_strategy = DatabaseStrategy(table_strategies=[
            TruncateTableStrategy("table1")
        ],
        after_scripts=["SELECT `after` FROM `after_table`;"]
    )
    provider.anonymize_database(database_strategy)

    truncate_call_index = manager.mock_calls.index(call.execution.MySqlCmdRunner().db_execute(query_factory.get_truncate_table.return_value))
    after_script_call_index = manager.mock_calls.index( call.execution.MySqlCmdRunner().db_execute("SELECT `after` FROM `after_table`;") )

    # after script should be called after anonymization
    assert truncate_call_index < after_script_call_index


@patch("pynonymizer.database.mysql.execution")
@patch("pynonymizer.database.mysql.query_factory")
def test_anonymize_database(query_factory, execution, simple_strategy, simple_strategy_update_fake_column, simple_strategy_update):
    manager = Mock()
    manager.attach_mock(execution, "execution")

    provider = MySqlProvider("1.2.3.4", "root", "password", "db_name")

    provider.anonymize_database(simple_strategy)

    seed_table_qualifier_map = ["_pynonymizer_seed_fake_data", {"user_name": simple_strategy_update_fake_column}]

    # get_create_seed_table( seed table name, qualifier map)
    query_factory.get_create_seed_table.assert_called_once_with(*seed_table_qualifier_map)

    # seed table should be called with the qual map 150 times
    query_factory.get_insert_seed_row.assert_has_calls( ([call(*seed_table_qualifier_map)] * 150) )

    query_factory.get_truncate_table.assert_called_once_with("truncate_table")
    query_factory.get_update_table.assert_called_once_with("_pynonymizer_seed_fake_data", simple_strategy_update)
    query_factory.get_drop_seed_table.assert_called_once_with("_pynonymizer_seed_fake_data")

    # anonymize basic order checks
    ix_create_seed = manager.mock_calls.index(call.execution.MySqlCmdRunner().db_execute(query_factory.get_create_seed_table()))
    ix_insert_seed_first = manager.mock_calls.index(call.execution.MySqlCmdRunner().db_execute(query_factory.get_insert_seed_row()))
    ix_insert_seed_last =  list_rindex(manager.mock_calls, call.execution.MySqlCmdRunner().db_execute(query_factory.get_insert_seed_row()))
    ix_trunc_table = manager.mock_calls.index(call.execution.MySqlCmdRunner().db_execute(query_factory.get_truncate_table()))
    ix_update_table = manager.mock_calls.index(call.execution.MySqlCmdRunner().db_execute(query_factory.get_update_table()))
    ix_drop_seed = manager.mock_calls.index(call.execution.MySqlCmdRunner().db_execute(query_factory.get_drop_seed_table()))

    assert ix_create_seed < ix_insert_seed_first
    assert ix_insert_seed_last < ix_trunc_table
    assert ix_insert_seed_last < ix_update_table
    assert ix_update_table < ix_drop_seed


import pytest
import unittest
from unittest.mock import patch, Mock
from pynonymizer.database.exceptions import DependencyError
from pynonymizer.database.postgres.execution import PSqlCmdRunner, PSqlDumpRunner
from tests.helpers import SuperdictOf
import subprocess


@patch("shutil.which", Mock(return_value=None))
class NoExecutablesInPathTests(unittest.TestCase):
    def test_dump_runner_missing_mysqldump(self):
        with pytest.raises(DependencyError):
            PSqlDumpRunner("1.2.3.4", "user", "password", "name")


    def test_cmd_runner_missing_mysql(self):
        with pytest.raises(DependencyError):
            PSqlCmdRunner("1.2.3.4", "user", "password", "name")


@patch("subprocess.Popen")
@patch("subprocess.check_output")
@patch("shutil.which", Mock(return_value="fake/path/to/executable"))
class DumperTests(unittest.TestCase):
    def test_open_dumper(self, check_output, popen):
        dump_runner = PSqlDumpRunner("1.2.3.4", "db_user", "db_password", "db_name")
        open_result = dump_runner.open_dumper()

        # dumper should open a process for the current db dump, piping stdout for processing
        popen.assert_called_with(['pg_dump', '--host', '1.2.3.4', '--username', 'db_user', 'db_name'],
                                 env=SuperdictOf({"PGPASSWORD": "db_password"}), stdout=subprocess.PIPE)

        # dumper should return the stdout of that process
        assert open_result == popen.return_value.stdout


@patch("subprocess.Popen")
@patch("subprocess.check_output")
@patch("shutil.which", Mock(return_value="fake/path/to/executable"))
class CmdTests(unittest.TestCase):
    def test_open_batch_processor(self, check_output, popen):
        cmd_runner = PSqlCmdRunner("1.2.3.4", "db_user", "db_password", "db_name")
        open_result = cmd_runner.open_batch_processor()

        # dumper should open a process for the current db dump, piping stdout for processing
        popen.assert_called_with(["psql", "--host", "1.2.3.4", "--username", "db_user", "--dbname", "db_name", "--quiet"],
                                 env=SuperdictOf({"PGPASSWORD": "db_password"}),
                                 stdin=subprocess.PIPE)

        # dumper should return the stdin of that process
        assert open_result == popen.return_value.stdin

    def test_execute(self, check_output, popen):
        """
        execute should execute an arbitrary statement with valid args
        """
        cmd_runner = PSqlCmdRunner("1.2.3.4", "db_user", "db_password", "db_name")
        execute_result = cmd_runner.execute("SELECT `column` from `table`;")

        check_output.assert_called_with(["psql", "--host", "1.2.3.4", "--username", "db_user", "--command",
                                         "SELECT `column` from `table`;"],
                                        env=SuperdictOf({"PGPASSWORD": "db_password"}),
                                        )

    def test_execute_list(self, check_output, popen):
        cmd_runner = PSqlCmdRunner("1.2.3.4", "db_user", "db_password", "db_name")
        execute_result = cmd_runner.execute(["SELECT `column` from `table`;", "SELECT `column2` from `table2`;"])

        check_output.assert_any_call(["psql", "--host", "1.2.3.4", "--username", "db_user", "--command",
                                         "SELECT `column` from `table`;"],
                                     env=SuperdictOf({"PGPASSWORD": "db_password"}),
                                     )

        check_output.assert_any_call(["psql", "--host", "1.2.3.4", "--username", "db_user", "--command",
                                         "SELECT `column2` from `table2`;"],
                                     env=SuperdictOf({"PGPASSWORD": "db_password"}),
                                     )

    def test_db_execute(self, check_output, popen):
        """
        execute should execute an arbitrary statement with valid args
        """
        cmd_runner = PSqlCmdRunner("1.2.3.4", "db_user", "db_password", "db_name")
        execute_result = cmd_runner.db_execute("SELECT `column` from `table`;")

        check_output.assert_called_with(["psql", "--host", "1.2.3.4", "--username", "db_user", "--dbname", "db_name", "--command", "SELECT `column` from `table`;"],
                                        env=SuperdictOf({"PGPASSWORD": "db_password"}),
                                        )

    def test_db_execute_list(self, check_output, popen):
        cmd_runner = PSqlCmdRunner("1.2.3.4", "db_user", "db_password", "db_name")
        execute_result = cmd_runner.db_execute(["SELECT `column` from `table`;", "SELECT `column2` from `table2`;"])

        check_output.assert_any_call(["psql", "--host", "1.2.3.4", "--username", "db_user", "--dbname", "db_name", "--command",
                                      "SELECT `column` from `table`;"],
                                     env=SuperdictOf({"PGPASSWORD": "db_password"}),
                                     )
        check_output.assert_any_call(["psql", "--host", "1.2.3.4", "--username", "db_user", "--dbname", "db_name", "--command",
                                      "SELECT `column2` from `table2`;"],
                                     env=SuperdictOf({"PGPASSWORD": "db_password"}),
                                     )

    def test_get_single_result(self, check_output, popen):
        """
        execute should execute an arbitrary statement and return the decoded, no-column result
        """
        cmd_runner = PSqlCmdRunner("1.2.3.4", "db_user", "db_password", "db_name")
        single_result = cmd_runner.get_single_result("SELECT `column` from `table`;")

        check_output.assert_called_with(["psql", "--host", "1.2.3.4", "--username", "db_user", "--dbname", "db_name", "-tA", "--command", "SELECT `column` from `table`;"],
                                        env=SuperdictOf({"PGPASSWORD": "db_password"}),
                                        )
        assert single_result == check_output.return_value.decode.return_value

    def test_connection_ok(self, check_output, popen):
        """
        When subprocess returns no errors, test should return true
        """
        cmd_runner = PSqlCmdRunner("1.2.3.4", "db_user", "db_password", "db_name")
        assert cmd_runner.test() is True

    def test_connection_bad(self, check_output, popen):
        """
        When subprocess throws calledProcessError, test should return false
        """
        check_output.side_effect = subprocess.CalledProcessError(1, "fakecmd")
        cmd_runner = PSqlCmdRunner("1.2.3.4", "db_user", "db_password", "db_name")
        assert cmd_runner.test() is False



import pytest
import unittest
from unittest.mock import patch, Mock, mock_open
from pynonymizer.__main__ import main
from pynonymizer.pynonymize import ArgumentValidationError, DatabaseConnectionError, pynonymize
from types import SimpleNamespace

def test_pynonymize_missing_db_credentials():
    with pytest.raises(ArgumentValidationError):
        pynonymize(
            input_path="input.sql",
            strategyfile_path="strategyfile.yml",
            output_path="output.sql",
            db_user=None,
            db_password=None
        )

@patch("dotenv.find_dotenv")
@patch("dotenv.load_dotenv")
@patch("pynonymizer.__main__.create_parser")
@patch("pynonymizer.__main__.pynonymize", autospec=True)
class MainArgTests(unittest.TestCase):
    def setUp(self):
        self.parsed_args = SimpleNamespace(
            legacy_input=None,
            legacy_strategyfile=None,
            legacy_output=None,
            input="TEST_INPUT",
            strategyfile="TEST_STRATEGYFILE",
            output="TEST_OUTPUT",
            db_type="TEST_TYPE",
            db_host="TEST_HOST",
            db_name="TEST_NAME",
            db_user="TEST_USER",
            db_password="TEST_PASSWORD",
            fake_locale="TEST_LOCALE",
            start_at_step="TEST_START_AT_STEP",
            skip_steps=["TEST_SKIP_1", "TEST_SKIP_2"],
            stop_at_step="TEST_STOP_AT_STEP",
            seed_rows=None,
            mssql_backup_compression=False
        )
    def test_dotenv_called(self, pynonymize, create_parser, load_dotenv, find_dotenv):
        """
        dotenv should be called
        """
        parser_mock = Mock(parse_args=Mock(return_value=self.parsed_args))
        create_parser.return_value = parser_mock

        main([])

        find_dotenv.assert_called()
        load_dotenv.assert_called()

    def test_arg_pass_legacy_override(self, pynonymize, create_parser, load_dotenv, find_dotenv):
        """
        the parsed set of args should be passed to the pynonymize main function
        legacy args should override normal ones to account for old positional behaviour
        """
        self.parsed_args.legacy_input = "LEGACY_INPUT"
        self.parsed_args.legacy_strategyfile = "LEGACY_STRATEGYFILE"
        self.parsed_args.legacy_output = "LEGACY_OUTPUT"
        parser_mock = Mock(parse_args=Mock(return_value=self.parsed_args))
        create_parser.return_value = parser_mock

        main([])

        create_parser.assert_called()
        parser_mock.parse_args.assert_called()
        pynonymize.assert_called_once_with(
            input_path="LEGACY_INPUT",
            strategyfile_path="LEGACY_STRATEGYFILE",
            output_path="LEGACY_OUTPUT",
            db_type="TEST_TYPE",
            db_host="TEST_HOST",
            db_name="TEST_NAME",
            db_user="TEST_USER",
            db_password="TEST_PASSWORD",
            fake_locale="TEST_LOCALE",
            start_at_step="TEST_START_AT_STEP",
            skip_steps=["TEST_SKIP_1", "TEST_SKIP_2"],
            stop_at_step="TEST_STOP_AT_STEP",
            seed_rows=None,
            mssql_backup_compression=False
        )

    def test_arg_pass_normal(self, pynonymize, create_parser, load_dotenv, find_dotenv):
        """
        the parsed set of args should be passed to the pynonymize main function
        """
        parser_mock = Mock(parse_args=Mock(return_value=self.parsed_args))
        create_parser.return_value = parser_mock

        main([])

        create_parser.assert_called()
        parser_mock.parse_args.assert_called()
        pynonymize.assert_called_once_with(
            input_path="TEST_INPUT",
            strategyfile_path="TEST_STRATEGYFILE",
            output_path="TEST_OUTPUT",
            db_type="TEST_TYPE",
            db_host="TEST_HOST",
            db_name="TEST_NAME",
            db_user="TEST_USER",
            db_password="TEST_PASSWORD",
            fake_locale="TEST_LOCALE",
            start_at_step="TEST_START_AT_STEP",
            skip_steps=["TEST_SKIP_1", "TEST_SKIP_2"],
            stop_at_step="TEST_STOP_AT_STEP",
            seed_rows=None,
            mssql_backup_compression=False
        )


@patch("dotenv.find_dotenv", Mock())
@patch("dotenv.load_dotenv", Mock())
@patch("pynonymizer.__main__.create_parser", Mock())
@patch("pynonymizer.__main__.pynonymize", Mock(side_effect=ArgumentValidationError(["test validation"])))
def test_sysexit_on_argument_invalid():
    """
    If pynonymize throws an argument validation error, main should exit with err 2
    """
    with pytest.raises(SystemExit) as e_info:
        main(["blah"])

    assert e_info.value.code == 2

@patch("dotenv.find_dotenv", Mock())
@patch("dotenv.load_dotenv", Mock())
@patch("pynonymizer.__main__.create_parser", Mock())
@patch("pynonymizer.__main__.pynonymize", Mock(side_effect=DatabaseConnectionError()))
def test_sysexit_on_database_connection_error():
    """
    If pynonymize throws an argument validation error, main should exit with err 1
    """
    with pytest.raises(SystemExit) as e_info:
        main(["blah"])

    assert e_info.value.code == 1


@patch("dotenv.find_dotenv", Mock())
@patch("dotenv.load_dotenv", Mock())
@patch("pynonymizer.pynonymize.read_config")
@patch("pynonymizer.pynonymize.get_provider")
@patch("pynonymizer.pynonymize.FakeColumnGenerator")
@patch("pynonymizer.pynonymize.StrategyParser")
@patch("builtins.open", mock_open(read_data="TESTFILEDATA"))
class MainProcessTests(unittest.TestCase):
    def test_any_db_kwarg(self, StrategyParser, FakeColumnSet, get_provider, read_config):
        """
        test that dynamic args are passed to the provider properly e.g. mssql_blah
        """
        pynonymize(
            input_path="TEST_INPUT",
            strategyfile_path="TEST_STRATEGYFILE",
            output_path="TEST_OUTPUT",
            db_type="mssql",
            db_host="TEST_HOST",
            db_name="TEST_NAME",
            db_user="TEST_USER",
            db_password="TEST_PASSWORD",
            fake_locale="TEST_LOCALE",
            mysql_other_amazing_var="TEST_DYNAMIC_VAR", # as this is mssql, this should be ignored
            mssql_special_provider_var="TEST_DYNAMIC_VAR2"
        )
        StrategyParser.return_value.parse_config.assert_called()
        get_provider.assert_called_with(type="mssql", db_host="TEST_HOST", db_user="TEST_USER", db_pass="TEST_PASSWORD", db_name="TEST_NAME", seed_rows=150, special_provider_var="TEST_DYNAMIC_VAR2")

    def test_pynonymize_main_process(self, StrategyParser, FakeColumnSet, get_provider, read_config):
        """
        a rough smoke test for the main process. This needs an integration test to back it up.
        """
        pynonymize(
            input_path="TEST_INPUT",
            strategyfile_path="TEST_STRATEGYFILE",
            output_path="TEST_OUTPUT",
            db_type="TEST_TYPE",
            db_host="TEST_HOST",
            db_name="TEST_NAME",
            db_user="TEST_USER",
            db_password="TEST_PASSWORD",
            fake_locale="TEST_LOCALE",
            seed_rows=999
        )
        StrategyParser.return_value.parse_config.assert_called()
        get_provider.assert_called_with(type="TEST_TYPE", db_host="TEST_HOST", db_user="TEST_USER", db_pass="TEST_PASSWORD", db_name="TEST_NAME", seed_rows=999)

        provider = get_provider.return_value
        provider.test_connection.assert_called()
        provider.create_database.assert_called()
        provider.restore_database.assert_called()
        provider.anonymize_database.assert_called()
        provider.dump_database.assert_called()
        provider.drop_database.assert_called()

    def test_pynonymize_stop_at_step(self, StrategyParser, FakeColumnSet, get_provider, read_config):
        pynonymize(
            input_path="TEST_INPUT",
            strategyfile_path="TEST_STRATEGYFILE",
            output_path="TEST_OUTPUT",
            db_type="TEST_TYPE",
            db_host="TEST_HOST",
            db_name="TEST_NAME",
            db_user="TEST_USER",
            db_password="TEST_PASSWORD",
            fake_locale="TEST_LOCALE",
            stop_at_step="ANONYMIZE_DB"
        )
        StrategyParser.return_value.parse_config.assert_called()
        get_provider.assert_called_with(type="TEST_TYPE", db_host="TEST_HOST", db_user="TEST_USER", db_pass="TEST_PASSWORD", db_name="TEST_NAME", seed_rows=150)

        provider = get_provider.return_value
        provider.test_connection.assert_called()
        provider.create_database.assert_called()
        provider.restore_database.assert_called()
        provider.anonymize_database.assert_called()
        provider.dump_database.assert_not_called()
        provider.drop_database.assert_not_called()

    def test_pynonymize_skip_steps(self, StrategyParser, FakeColumnSet, get_provider,
                                     read_config):
        pynonymize(
            input_path="TEST_INPUT",
            strategyfile_path="TEST_STRATEGYFILE",
            output_path="TEST_OUTPUT",
            db_type="TEST_TYPE",
            db_host="TEST_HOST",
            db_name="TEST_NAME",
            db_user="TEST_USER",
            db_password="TEST_PASSWORD",
            fake_locale="TEST_LOCALE",
            skip_steps=["ANONYMIZE_DB", "CREATE_DB", "DUMP_DB"]
        )
        get_provider.assert_called_with(type="TEST_TYPE", db_host="TEST_HOST", db_user="TEST_USER", db_pass="TEST_PASSWORD", db_name="TEST_NAME", seed_rows=150)

        provider = get_provider.return_value
        provider.test_connection.assert_called()
        provider.create_database.assert_not_called()
        provider.restore_database.assert_called()
        provider.anonymize_database.assert_not_called()
        provider.dump_database.assert_not_called()
        provider.drop_database.assert_called()

    def test_pynonymize_start_at_step(self, StrategyParser, FakeColumnSet, get_provider,
                                      read_config):
        pynonymize(
            input_path="TEST_INPUT",
            strategyfile_path="TEST_STRATEGYFILE",
            output_path="TEST_OUTPUT",
            db_type="TEST_TYPE",
            db_host="TEST_HOST",
            db_name="TEST_NAME",
            db_user="TEST_USER",
            db_password="TEST_PASSWORD",
            fake_locale="TEST_LOCALE",
            start_at_step="ANONYMIZE_DB"
        )
        StrategyParser.return_value.parse_config.assert_called()
        get_provider.assert_called_with(type="TEST_TYPE", db_host="TEST_HOST", db_user="TEST_USER", db_pass="TEST_PASSWORD", db_name="TEST_NAME", seed_rows=150)

        provider = get_provider.return_value
        provider.test_connection.assert_called()
        provider.create_database.assert_not_called()
        provider.restore_database.assert_not_called()
        provider.anonymize_database.assert_called()
        provider.dump_database.assert_called()
        provider.drop_database.assert_called()


@patch("dotenv.find_dotenv", Mock())
@patch("dotenv.load_dotenv", Mock())
@patch("pynonymizer.pynonymize.read_config")
@patch("pynonymizer.pynonymize.get_provider")
@patch("pynonymizer.pynonymize.FakeColumnGenerator")
@patch("pynonymizer.pynonymize.StrategyParser")
@patch("builtins.open", mock_open(read_data="TESTFILEDATA"))
class OptionalArgumentsSkippedTests(unittest.TestCase):
    """
    pynonymize should not throw argument validation errors for missing "mandatory" args
    that are only mandatory for certain steps.

    START = 0
    GET_SOURCE = 100
    CREATE_DB = 200
    RESTORE_DB = 300
    ANONYMIZE_DB = 400
    DUMP_DB = 500
    DROP_DB = 600
    END = 9999

    """
    def test_optional_input_when_skip_input_steps(self, StrategyParser, FakeColumnSet,
                                                  get_provider, read_config):
        pynonymize(
            input_path=None,
            strategyfile_path="TEST_STRATEGYFILE",
            output_path="TEST_OUTPUT",
            db_type="TEST_TYPE",
            db_host="TEST_HOST",
            db_name="TEST_NAME",
            db_user="TEST_USER",
            db_password="TEST_PASSWORD",
            fake_locale="TEST_LOCALE",
            start_at_step=None,
            stop_at_step=None,
            skip_steps=["RESTORE_DB"]
        )

        StrategyParser.return_value.parse_config.assert_called()
        get_provider.assert_called_with(type="TEST_TYPE", db_host="TEST_HOST", db_user="TEST_USER", db_pass="TEST_PASSWORD", db_name="TEST_NAME", seed_rows=150)

        provider = get_provider.return_value
        provider.test_connection.assert_called()
        provider.create_database.assert_called()
        provider.restore_database.assert_not_called()
        provider.anonymize_database.assert_called()
        provider.dump_database.assert_called()
        provider.drop_database.assert_called()

    def test_optional_input_when_start_at_after_input_steps(self, StrategyParser, FakeColumnSet,
                                                  get_provider, read_config):
        pynonymize(
            input_path=None,
            strategyfile_path="TEST_STRATEGYFILE",
            output_path="TEST_OUTPUT",
            db_type="TEST_TYPE",
            db_host="TEST_HOST",
            db_name="TEST_NAME",
            db_user="TEST_USER",
            db_password="TEST_PASSWORD",
            fake_locale="TEST_LOCALE",
            start_at_step="ANONYMIZE_DB",
            stop_at_step=None,
            skip_steps=None
        )

        StrategyParser.return_value.parse_config.assert_called()
        get_provider.assert_called_with(type="TEST_TYPE", db_host="TEST_HOST", db_user="TEST_USER", db_pass="TEST_PASSWORD", db_name="TEST_NAME", seed_rows=150)

        provider = get_provider.return_value
        provider.test_connection.assert_called()
        provider.create_database.assert_not_called()
        provider.restore_database.assert_not_called()
        provider.anonymize_database.assert_called()
        provider.dump_database.assert_called()
        provider.drop_database.assert_called()

    def test_optional_input_when_stop_at_before_input_steps(self, StrategyParser, FakeColumnSet,
                                                  get_provider, read_config):
        pynonymize(
            input_path=None,
            strategyfile_path="TEST_STRATEGYFILE",
            output_path="TEST_OUTPUT",
            db_type="TEST_TYPE",
            db_host="TEST_HOST",
            db_name="TEST_NAME",
            db_user="TEST_USER",
            db_password="TEST_PASSWORD",
            fake_locale="TEST_LOCALE",
            start_at_step=None,
            stop_at_step="CREATE_DB",
            skip_steps=None
        )

        get_provider.assert_called_with(type="TEST_TYPE", db_host="TEST_HOST", db_user="TEST_USER", db_pass="TEST_PASSWORD", db_name="TEST_NAME", seed_rows=150)

        provider = get_provider.return_value
        provider.test_connection.assert_called()
        provider.create_database.assert_called()
        provider.restore_database.assert_not_called()
        provider.anonymize_database.assert_not_called()
        provider.dump_database.assert_not_called()
        provider.drop_database.assert_not_called()

    def test_optional_strategyfile_when_skip_anonymize(self, StrategyParser, FakeColumnSet,
                                                  get_provider, read_config):
        pynonymize(
            input_path="TEST_INPUT",
            strategyfile_path=None,
            output_path="TEST_OUTPUT",
            db_type="TEST_TYPE",
            db_host="TEST_HOST",
            db_name="TEST_NAME",
            db_user="TEST_USER",
            db_password="TEST_PASSWORD",
            fake_locale="TEST_LOCALE",
            start_at_step=None,
            stop_at_step=None,
            skip_steps=["ANONYMIZE_DB"]
        )
        get_provider.assert_called_with(type="TEST_TYPE", db_host="TEST_HOST", db_user="TEST_USER", db_pass="TEST_PASSWORD", db_name="TEST_NAME", seed_rows=150)

        provider = get_provider.return_value
        provider.test_connection.assert_called()
        provider.create_database.assert_called()
        provider.restore_database.assert_called()
        provider.anonymize_database.assert_not_called()
        provider.dump_database.assert_called()
        provider.drop_database.assert_called()

    def test_optional_strategyfile_when_start_at_after_anonymize(self, StrategyParser, FakeColumnSet,
                                                  get_provider, read_config):
        pynonymize(
            input_path="TEST_INPUT",
            strategyfile_path=None,
            output_path="TEST_OUTPUT",
            db_type="TEST_TYPE",
            db_host="TEST_HOST",
            db_name="TEST_NAME",
            db_user="TEST_USER",
            db_password="TEST_PASSWORD",
            fake_locale="TEST_LOCALE",
            start_at_step="DUMP_DB",
            stop_at_step=None,
            skip_steps=None
        )
        get_provider.assert_called_with(type="TEST_TYPE", db_host="TEST_HOST", db_user="TEST_USER", db_pass="TEST_PASSWORD", db_name="TEST_NAME", seed_rows=150)

        provider = get_provider.return_value
        provider.test_connection.assert_called()
        provider.create_database.assert_not_called()
        provider.restore_database.assert_not_called()
        provider.anonymize_database.assert_not_called()
        provider.dump_database.assert_called()
        provider.drop_database.assert_called()

    def test_optional_strategyfile_when_stop_at_before_anonymize(self, StrategyParser, FakeColumnSet,
                                                  get_provider, read_config):
        pynonymize(
            input_path="TEST_INPUT",
            strategyfile_path=None,
            output_path="TEST_OUTPUT",
            db_type="TEST_TYPE",
            db_host="TEST_HOST",
            db_name="TEST_NAME",
            db_user="TEST_USER",
            db_password="TEST_PASSWORD",
            fake_locale="TEST_LOCALE",
            start_at_step=None,
            stop_at_step="RESTORE_DB",
            skip_steps=None
        )
        get_provider.assert_called_with(type="TEST_TYPE", db_host="TEST_HOST", db_user="TEST_USER", db_pass="TEST_PASSWORD", db_name="TEST_NAME", seed_rows=150)

        provider = get_provider.return_value
        provider.test_connection.assert_called()
        provider.create_database.assert_called()
        provider.restore_database.assert_called()
        provider.anonymize_database.assert_not_called()
        provider.dump_database.assert_not_called()
        provider.drop_database.assert_not_called()

    def test_optional_output_when_skip_output_steps(self, StrategyParser, FakeColumnSet,
                                                  get_provider, read_config):
        pynonymize(
            input_path="TEST_INPUT",
            strategyfile_path="TEST_STRATEGYFILE",
            output_path=None,
            db_type="TEST_TYPE",
            db_host="TEST_HOST",
            db_name="TEST_NAME",
            db_user="TEST_USER",
            db_password="TEST_PASSWORD",
            fake_locale="TEST_LOCALE",
            start_at_step=None,
            stop_at_step=None,
            skip_steps=["DUMP_DB"]
        )

        StrategyParser.return_value.parse_config.assert_called()
        get_provider.assert_called_with(type="TEST_TYPE", db_host="TEST_HOST", db_user="TEST_USER", db_pass="TEST_PASSWORD", db_name="TEST_NAME", seed_rows=150)

        provider = get_provider.return_value
        provider.test_connection.assert_called()
        provider.create_database.assert_called()
        provider.restore_database.assert_called()
        provider.anonymize_database.assert_called()
        provider.dump_database.assert_not_called()
        provider.drop_database.assert_called()

    def test_optional_output_when_start_at_after_output_steps(self, StrategyParser, FakeColumnSet,
                                                  get_provider, read_config):
        pynonymize(
            input_path="TEST_INPUT",
            strategyfile_path="TEST_STRATEGYFILE",
            output_path=None,
            db_type="TEST_TYPE",
            db_host="TEST_HOST",
            db_name="TEST_NAME",
            db_user="TEST_USER",
            db_password="TEST_PASSWORD",
            fake_locale="TEST_LOCALE",
            start_at_step="DROP_DB",
            stop_at_step=None,
            skip_steps=None
        )
        get_provider.assert_called_with(type="TEST_TYPE", db_host="TEST_HOST", db_user="TEST_USER", db_pass="TEST_PASSWORD", db_name="TEST_NAME", seed_rows=150)

        provider = get_provider.return_value
        provider.test_connection.assert_called()
        provider.create_database.assert_not_called()
        provider.restore_database.assert_not_called()
        provider.anonymize_database.assert_not_called()
        provider.dump_database.assert_not_called()
        provider.drop_database.assert_called()

    def test_optional_output_when_stop_at_before_output_steps(self, StrategyParser, FakeColumnSet,
                                                  get_provider, read_config):
        pynonymize(
            input_path="TEST_INPUT",
            strategyfile_path="TEST_STRATEGYFILE",
            output_path=None,
            db_type="TEST_TYPE",
            db_host="TEST_HOST",
            db_name="TEST_NAME",
            db_user="TEST_USER",
            db_password="TEST_PASSWORD",
            fake_locale="TEST_LOCALE",
            start_at_step=None,
            stop_at_step=None,
            skip_steps=["DUMP_DB"]
        )

        StrategyParser.return_value.parse_config.assert_called()
        get_provider.assert_called_with(type="TEST_TYPE", db_host="TEST_HOST", db_user="TEST_USER", db_pass="TEST_PASSWORD", db_name="TEST_NAME", seed_rows=150)

        provider = get_provider.return_value
        provider.test_connection.assert_called()
        provider.create_database.assert_called()
        provider.restore_database.assert_called()
        provider.anonymize_database.assert_called()
        provider.dump_database.assert_not_called()
        provider.drop_database.assert_called()



import pytest
from unittest import TestCase
from unittest.mock import mock_open, patch
from pynonymizer.database.basic.input import GzipInput, RawInput, UnknownInputTypeError, resolve_input


def test_gzip_open():
    with patch("gzip.open", mock_open(read_data=b"data")) as mock_file:
        gz = GzipInput("testfile.gz")
        open_result = gz.open()

        mock_file.assert_called_with("testfile.gz", "rb")
        assert open_result == mock_file.return_value


def test_gzip_get_size():
    with patch("builtins.open", mock_open(read_data=b'X\x1f\x00\x00')) as mock_file:
        gz = GzipInput("testfile.gz")
        assert gz.get_size() == 8024


def test_raw_open():
    """
    rawfile should return the file object directly
    """
    with patch("builtins.open", mock_open(read_data=b"data")) as mock_file:
        raw = RawInput("testfile.sql")
        open_result = raw.open()

        assert open_result == mock_file.return_value


@patch("os.path.getsize", autospec=True)
def test_raw_get_size(getsize):
    """
    rawfile should use os.path.getsize to bring back a size
    """
    raw = RawInput("testfile.sql")

    size_result = raw.get_size()
    assert size_result == getsize.return_value


class ResolveFromFilepathTest(TestCase):
    test_path_examples = [
        "",
        "complex/multi/part/path/test/",
        "complex\\multi\\part\\path\\test\\",
        "/absolute/path/test/",
        "C:\\absolute\\path\\test\\"
    ]

    def test_resolve_raw(self):
        for path in self.test_path_examples:
            test_path = path + "test.sql"
            with self.subTest(i=test_path):
                assert isinstance(resolve_input(test_path), RawInput)

    def test_resolve_gzip(self):
        for path in self.test_path_examples:
            test_path = path + "test.sql.gz"
            with self.subTest(i=test_path):
                assert isinstance(resolve_input(test_path), GzipInput)

    def test_resolve_unknown(self):
        with pytest.raises(UnknownInputTypeError):
            resolve_input("unknown_file.bbs")


from pynonymizer.database.mssql import MsSqlProvider
from pynonymizer.database.exceptions import DependencyError
import pytest
import pyodbc
from decimal import Decimal
from unittest.mock import patch, Mock, call
from tests.helpers import list_rindex


@pytest.fixture
def provider():
    return MsSqlProvider(None, "DB_USER", "DB_PASS", "DB_NAME")

@pytest.fixture
def provider_with_compression():
    return MsSqlProvider(None, "DB_USER", "DB_PASS", "DB_NAME", backup_compression=True)


def mock_restore_side_effect(statement, *args, **kwargs):
    if statement.strip().startswith("RESTORE FILELISTONLY"):
        filelist = Mock()
        filelist.fetchall.return_value = [('AdventureWorks2016_EXT_Data', 'C:\\Program Files\\Microsoft SQL Server\\MSSQL13.MSSQL2016RTM\\MSSQL\\DATA\\AdventureWorks2016_EXT_Data.mdf', 'D', 'PRIMARY', 405078016, 35184372080640, 1, Decimal('0'), Decimal('0'), '78D8A94F-4CFF-4ED0-9846-3AF3BE2F1C7D', Decimal('0'), Decimal('0'), 394002432, 512, 1, None, Decimal('108000000652300212'), 'A7CE522E-427E-4375-8054-DA6B304B6C4F', False, True, None, None), ('AdventureWorks2016_EXT_Log', 'C:\\Program Files\\Microsoft SQL Server\\MSSQL13.MSSQL2016RTM\\MSSQL\\DATA\\AdventureWorks2016_EXT_Log.ldf', 'L', None, 186646528, 2199023255552, 2, Decimal('0'), Decimal('0'), '9B6653F1-815D-4064-82F6-C43D81CABD61', Decimal('0'), Decimal('0'), 0, 512, 0, None, Decimal('0'), '00000000-0000-0000-0000-000000000000', False, True, None, None), ('AdventureWorks2016_EXT_mod', 'C:\\Program Files\\Microsoft SQL Server\\MSSQL13.MSSQL2016RTM\\MSSQL\\DATA\\AdventureWorks2016_EXT_mod', 'S', 'AdventureWorks2016_EXT_mod', 0, 0, 65537, Decimal('43000002114900003'), Decimal('0'), 'E3B5DA3E-E186-491A-A723-E4A6852C0B55', Decimal('0'), Decimal('0'), 1025310720, 512, 2, None, Decimal('108000000652300212'), 'A7CE522E-427E-4375-8054-DA6B304B6C4F', False, True, None, None)]
        return filelist
    if statement.strip().startswith("RESTORE DATABASE"):
        restore = Mock()
        restore.nextset.side_effect = [True, True, False]
        return restore
    if statement.strip().startswith("SELECT"):
        selectvar = Mock()
        selectvar.fetchone.return_value = ("C:\\test.mdf",)
        return selectvar

    return Mock()


def mock_backup_side_effect(statement, *args, **kwargs):
    if statement.strip().startswith("BACKUP DATABASE"):
        backup_mock = Mock()
        backup_mock.nextset.side_effect = [True, True, False]
        return backup_mock

    return Mock()


def test_raise_on_remote_server():
    with pytest.raises(DependencyError):
        MsSqlProvider("192.168.2.1", "username", "password", "dbname")

@patch("pyodbc.connect")
def test_create_database_noop(connect, provider):
    """
    create_database should do nothing.
    """
    provider.create_database()

    connect.assert_not_called()
    connect.return_value.execute.assert_not_called()


@patch("pyodbc.connect")
def test_test_connection(connect, provider):
    good_connection = provider.test_connection()

    assert good_connection is True
    connect.return_value.execute.assert_called_with("SELECT @@VERSION;")


@patch("pyodbc.connect")
def test_test_connection_fail(connect, provider):
    connect.return_value.execute.side_effect = pyodbc.OperationalError()

    good_connection = provider.test_connection()

    assert good_connection is False
    connect.return_value.execute.assert_called_with("SELECT @@VERSION;")


@patch("pyodbc.connect")
def test_restore_database(connect, provider):
    connect.return_value.execute.side_effect = mock_restore_side_effect
    provider.restore_database("test.bak")

    connect.return_value.execute.assert_any_call(
        'RESTORE DATABASE ? FROM DISK = ? WITH MOVE ? TO ?, MOVE ? TO ?, MOVE ? TO ?, STATS = ?;',
        [
            'DB_NAME', 'test.bak', 'AdventureWorks2016_EXT_Data', 'C:\\DB_NAME_AdventureWorks2016_EXT_Data.mdf',
            'AdventureWorks2016_EXT_Log', 'C:\\DB_NAME_AdventureWorks2016_EXT_Log.ldf', 'AdventureWorks2016_EXT_mod',
            'C:\\DB_NAME_AdventureWorks2016_EXT_mod', 5
        ]
    )


@patch("pyodbc.connect")
def test_dump_database(connect, provider):
    connect.return_value.execute.side_effect = mock_backup_side_effect
    provider.dump_database("test_output.bak")

    connect.return_value.execute.assert_any_call("BACKUP DATABASE ? TO DISK = ? WITH STATS = ?;", ["DB_NAME", "test_output.bak", 5])

@patch("pyodbc.connect")
def test_dump_database_compression(connect, provider_with_compression):
    connect.return_value.execute.side_effect = mock_backup_side_effect
    provider_with_compression.dump_database("test_output.bak")

    connect.return_value.execute.assert_any_call("BACKUP DATABASE ? TO DISK = ? WITH COMPRESSION, STATS = ?;", ["DB_NAME", "test_output.bak", 5])


@patch("pyodbc.connect")
def test_drop_database(connect, provider):
    provider.drop_database()

    connect.return_value.execute.assert_any_call("DROP DATABASE IF EXISTS [DB_NAME];")


@patch("pyodbc.connect")
def test_anonymize(connect, provider, simple_strategy, simple_strategy_fake_generator):
    provider.anonymize_database(simple_strategy)

    execute_calls = connect().execute.mock_calls

    ix_create_seed = execute_calls.index(call('CREATE TABLE [_pynonymizer_seed_fake_data]([user_name] VARCHAR(MAX));'))
    ix_insert_seed_first = execute_calls.index(call('INSERT INTO [_pynonymizer_seed_fake_data]([user_name]) VALUES ( ?);', ['TEST_VALUE']))
    ix_insert_seed_last = list_rindex(execute_calls, call('INSERT INTO [_pynonymizer_seed_fake_data]([user_name]) VALUES ( ?);', ['TEST_VALUE']) )
    ix_trunc_table = execute_calls.index(call('TRUNCATE TABLE [truncate_table];'))
    ix_update_table_1 = execute_calls.index(call("UPDATE [update_table_where_3] SET [column1] = ( SELECT CONCAT(NEWID(), '@', NEWID(), '.com') ),[column2] = ( SELECT NEWID() ) WHERE BANANAS < 5;"))
    ix_update_table_2 = execute_calls.index(call('UPDATE [update_table_where_3] SET [column3] = ( SELECT TOP 1 [user_name] FROM [_pynonymizer_seed_fake_data] ORDER BY NEWID()) WHERE BANANAS < 3;'))
    ix_update_table_3 = execute_calls.index(call("UPDATE [update_table_where_3] SET [column4] = ('');"))
    ix_drop_seed = execute_calls.index(call('DROP TABLE IF EXISTS [_pynonymizer_seed_fake_data];'))

    # seed table create needs to happen before inserting data
    assert ix_create_seed < ix_insert_seed_first

    # last insert of seed data before update anonymize starts
    assert ix_insert_seed_last < ix_trunc_table
    assert ix_insert_seed_last < ix_update_table_1
    assert ix_insert_seed_last < ix_update_table_2
    assert ix_insert_seed_last < ix_update_table_3

    # update anonymize should all be before seed table drop
    assert ix_update_table_1 < ix_drop_seed
    assert ix_update_table_2 < ix_drop_seed
    assert ix_update_table_3 < ix_drop_seed


from datetime import date, datetime
from pynonymizer.database.exceptions import UnsupportedColumnStrategyError
from pynonymizer.strategy.update_column import UpdateColumnStrategyTypes
from pynonymizer.fake import FakeDataType
"""
All Static query generation functions
"""
_FAKE_COLUMN_TYPES = {
    FakeDataType.STRING: "VARCHAR(65535)",
    FakeDataType.DATE: "DATE",
    FakeDataType.DATETIME: "DATETIME",
    FakeDataType.INT: "INT"
}

# Random text function
_RAND_MD5 = "md5(random()::text)"


def _get_sql_type(data_type):
    return _FAKE_COLUMN_TYPES[data_type]


def _get_column_subquery(seed_table_name, column_strategy):
    if column_strategy.strategy_type == UpdateColumnStrategyTypes.EMPTY:
        return "('')"
    elif column_strategy.strategy_type == UpdateColumnStrategyTypes.UNIQUE_EMAIL:
        return f"( SELECT CONCAT({_RAND_MD5}, '@', {_RAND_MD5}, '.com') )"
    elif column_strategy.strategy_type == UpdateColumnStrategyTypes.UNIQUE_LOGIN:
        return f"( SELECT {_RAND_MD5} )"
    elif column_strategy.strategy_type == UpdateColumnStrategyTypes.FAKE_UPDATE:
        return f"( SELECT {column_strategy.qualifier} FROM {seed_table_name} ORDER BY RANDOM() LIMIT 1)"
    elif column_strategy.strategy_type == UpdateColumnStrategyTypes.LITERAL:
        return column_strategy.value
    else:
        raise UnsupportedColumnStrategyError(column_strategy)


def _escape_sql_value(value):
    """
    return a sql-ified version of a seed column's value
    Normally this defines the stringification of datatypes and escaping for strings
    """
    if isinstance(value, (str, datetime, date)):
        return "'" + str(value).replace("'", "''") + "'"
    else:
        return str(value)


def _get_qualified_table_name(schema, table):
    return f"{schema}.{table}" if schema else table


def get_truncate_table(table_strategy):
    return f"TRUNCATE TABLE {_get_qualified_table_name(table_strategy.schema, table_strategy.table_name)} CASCADE;"


def get_create_seed_table(table_name, qualifier_map):
    if len(qualifier_map) < 1:
        raise ValueError("Cannot create a seed table with no columns")

    create_columns = [f"{qualifier} {_get_sql_type(strategy.data_type)}" for qualifier, strategy in qualifier_map.items()]

    return "CREATE TABLE {} ({});".format(table_name, ",".join(create_columns) )


def get_drop_seed_table(table_name):
    return f"DROP TABLE IF EXISTS {table_name};"


def get_insert_seed_row(table_name, qualifier_map):

    column_names = ",".join( [f"{qualifier}" for qualifier in qualifier_map.keys()] )
    column_values = ",".join( [f"{_escape_sql_value(strategy.value)}" for strategy in qualifier_map.values()] )

    return "INSERT INTO {}({}) VALUES ({});".format(table_name, column_names, column_values)


def get_create_database(database_name):
    return f"CREATE DATABASE {database_name};"


def get_drop_database(database_name):
    return [
        # terminate other connections so we can drop
        f"SELECT pid, pg_terminate_backend(pid) FROM pg_stat_activity WHERE datname = '{database_name}' AND pid != pg_backend_pid();",
        f"DROP DATABASE IF EXISTS {database_name};"
    ]


def get_update_table(seed_table_name, update_table_strategy):
    # group on where_condition
    # build lists of update statements based on the where
    output_statements = []
    where_update_statements = {}
    for where, column_map in update_table_strategy.group_by_where().items():
        where_update_statements[where] = []
        for column_name, column_strategy in column_map.items():
            where_update_statements[where].append("{} = {}".format(
                column_name,
                _get_column_subquery(seed_table_name, column_strategy))
            )

        assignments = ",".join( where_update_statements[where] )
        where_clause = f" WHERE {where}" if where else ""

        output_statements.append(
            "UPDATE {} SET {}{};".format(
                _get_qualified_table_name(update_table_strategy.schema, update_table_strategy.table_name),
                assignments,
                where_clause
            )
        )

    return output_statements


# TODO: fix for postgres
def get_dumpsize_estimate(database_name):
    return "SELECT 1;"
